<?php

namespace Drupal\drupalorg_migrate\Drush\Commands;

use Drupal\Core\Database\Database;
use Drupal\drupalorg\ProjectService;
use Drupal\node\Entity\Node;
use Drush\Attributes as CLI;
use Drush\Commands\DrushCommands as BaseDrushCommands;

/**
 * A drush command file.
 *
 * @package Drupal\drupalorg_migrate\Commands
 */
final class DrushCommands extends BaseDrushCommands {

  /**
   * Size of the batch to use.
   *
   * Memory usage around 60MB for 100000.
   *
   * @var int
   */
  const BATCH_LENGTH = 50000;

  /**
   * Node types to check in addition to project types.
   *
   * @var array
   */
  const NODE_TYPES_TO_CHECK = [
    ...ProjectService::PROJECT_TYPES,
    'project_release',
    // 'casestudy',
  ];

  /**
   * Check users deleted in D7 and not in D10.
   */
  #[CLI\Command(name: 'drupalorg_migrate:sync-users')]
  public function syncUsers() {
    $this->deleteEntitiesNotComingFromMigration('users');
  }

  /**
   * Check projects deleted in D7 and not in D10.
   */
  #[CLI\Command(name: 'drupalorg_migrate:sync-projects')]
  public function syncProjects() {
    $this->deleteEntitiesNotComingFromMigration('node');
  }

  /**
   * Check entities in D7 and not in D10.
   */
  #[CLI\Command(name: 'drupalorg_migrate:check-entities')]
  public function checkEntities() {
    $this->checkEntitiesNotMigrated('users');
    $this->checkEntitiesNotMigrated('node');
  }

  /**
   * Check entities that are not migrated.
   *
   * @param string $type
   *   Type of entities to check.
   */
  protected function checkEntitiesNotMigrated($type) {
    $bundles = [];
    if ($type === 'node') {
      $bundles = self::NODE_TYPES_TO_CHECK;
    }
    elseif ($type === 'users') {
      // Special case as the user has no bundles.
      $bundles = [NULL];
    }
    foreach ($bundles as $entity_type) {
      $this->io()->info(\dt('Checking @type (@entity_type) entities migration status.', [
        '@type' => $type,
        '@entity_type' => $entity_type,
      ]));
      $entities_not_migrated = $this->entitiesFromFirstDatabaseNotInSecondDatabase('migrate', 'default', $type, $entity_type);
      if (empty($entities_not_migrated)) {
        $this->io()->success(\dt('Could not find any @type (@entity_type) entity not migrated.', [
          '@type' => $type,
          '@entity_type' => $entity_type,
        ]));
      }
      else {
        $count_entities_not_migrated = count($entities_not_migrated);
        $entity_ids_as_string = $this->arrayToStringTruncated($entities_not_migrated);
        $this->io()->warning(\dt('@count @type (@entity_type) entities are not migrated, with the following IDs: @ids', [
          '@count' => $count_entities_not_migrated,
          '@type' => $type,
          '@entity_type' => $entity_type,
          '@ids' => $entity_ids_as_string,
        ]));
      }
    }
  }

  /**
   * Remove entities that are not coming from migrations.
   *
   * @param string $type
   *    Type of entities to check.
   */
  protected function deleteEntitiesNotComingFromMigration($type) {
    $bundles = [];
    if ($type === 'node') {
      $bundles = self::NODE_TYPES_TO_CHECK;
    }
    elseif ($type === 'users') {
      // Special case as the user has no bundles.
      $bundles = [NULL];
    }
    foreach ($bundles as $entity_type) {
      $this->io()->info(\dt('Checking @type (@entity_type) that are not coming from a migration.', [
        '@type' => $type,
        '@entity_type' => $entity_type,
      ]));
      $extra_entities = $this->entitiesFromFirstDatabaseNotInSecondDatabase('default', 'migrate', $type, $entity_type);
      if (empty($extra_entities)) {
        $this->io()->success(\dt('Could not find any @type (@entity_type) to delete.', [
          '@type' => $type,
          '@entity_type' => $entity_type,
        ]));
      }
      else {
        $count_entities_to_remove = count($extra_entities);
        $entities_to_remove = $this->arrayToStringTruncated($extra_entities);
        $warning = \dt('@count @type (@entity_type) entities will be deleted, with the following IDs: @ids', [
          '@type' => $type,
          '@entity_type' => $entity_type,
          '@count' => $count_entities_to_remove,
          '@ids' => $entities_to_remove,
        ]);
        $this->io()->warning($warning);

        if (
          $count_entities_to_remove &&
          $this->io()->confirm(dt('Do you want to delete the @type (@entity_type) entities?', [
            '@type' => $type,
            '@entity_type' => $entity_type,
          ]))
        ) {
          $this->logger()->warning($warning);
          if ($count_entities_to_remove > 200) {
            // In this case, there is something very wrong with the data. Do not delete.
            $error = dt('Number of entities to be deleted (@count) exceeded the limit (@limit)', [
              '@count' => $count_entities_to_remove,
              '@limit' => 200,
            ]);
            $this->io()->error($error);
            $this->logger()->error($error);
          }
          else {
            foreach ($extra_entities as $extra_entity_id) {
              if ($type === 'users') {
                $this->cancelAccount($extra_entity_id);
              }
              elseif ($type === 'node') {
                $this->deleteNode($extra_entity_id);
              }
            }
          }
        }
      }
    }
  }

  /**
   * Show array as string (comma separated) and truncate it if too long.
   *
   * @param array $array
   *   Array to convert to string.
   *
   * @return string
   *   String representation of the array.
   */
  protected function arrayToStringTruncated($array) {
    return (count($array) > 100) ?
      implode(', ', array_slice($array, 0, 100)) . '... (truncated)' :
      implode(', ', $array);
  }

  /**
   * Find entity ids present in a database and not in another.
   *
   * @param string $first_database
   *   Key of the database.
   * @param string $second_database
   *   Key of the database.
   * @param string $entity_type_id
   *   Type of the entity.
   * @param string $bundle
   *   (Optional) Bundle of the entity.
   *
   * @return array
   *   List of entity ids.
   */
  protected function entitiesFromFirstDatabaseNotInSecondDatabase($first_database, $second_database, $entity_type_id, $bundle = NULL) {
    $count_entities = $this->getEntityCounts($first_database, $entity_type_id, $bundle);
    $length = self::BATCH_LENGTH;
    $offset = 0;
    $extra_entities = [];

    $ids = $this->getNextEntityBatch($offset, $length, $first_database, $entity_type_id, $bundle);
    $this->io()->progressStart($count_entities);
    while (!empty($ids)) {
      $ids_not_in_database = $this->entityIdsNotInDatabase($ids, $second_database, $entity_type_id, $bundle);
      $extra_entities = array_merge($extra_entities, $ids_not_in_database);

      $this->io()->progressAdvance($length);

      $offset += $length;
      $ids = $this->getNextEntityBatch($offset, $length, $first_database, $entity_type_id, $bundle);
    }
    $this->io()->progressFinish();

    return $extra_entities;
  }

  /**
   * Deletes a node.
   *
   * @param int $nid
   *   The node id.
   */
  protected function deleteNode($nid) {
    $node = Node::load($nid);
    if ($node) {
      $node->delete();
      $this->io()->info(dt('Node @nid deleted.', [
        '@nid' => $nid,
      ]));
    }
  }

  /**
   * Cancels a user account.
   *
   * @param int $uid
   *   User id.
   */
  protected function cancelAccount($uid) {
    // No need for extra logging as the user_cancel function does it.
    user_cancel([], $uid, 'user_cancel_reassign');
    $batch =& batch_get();
    $batch['progressive'] = FALSE;
    // Batch it the drush way, instead of batch_process().
    drush_backend_batch_process();
  }

  /**
   * Gets a count of entity table on the given database for a given type.
   *
   * @param string $database
   *   Database to check.
   * @param string $entity_type_id
   *   Type of the entity.
   * @param string $bundle
   *   (Optional) Type of the node.
   *
   * @return int
   *   Count of entities.
   */
  protected function getEntityCounts($database, $entity_type_id, $bundle = NULL) {
    $connection = Database::getConnection('default', $database);
    $query = $connection->select($entity_type_id, 'e');
    if (!is_null($bundle)) {
      $query->condition('e.type', $bundle);
    }
    return $query
      ->countQuery()
      ->execute()
      ->fetchField();
  }

  /**
   * Get a batch of entity ids to check.
   *
   * @param int $offset
   *   Offset of the records to get.
   * @param int $length
   *   Length of the records to get.
   * @param string $database
   *   Database to check.
   * @param string $entity_type_id
   *   Type of the entity.
   * @param string $bundle
   *   (Optional) Bundle types to bring.
   *
   * @return array
   *   Array of entity ids.
   */
  protected function getNextEntityBatch($offset, $length, $database, $entity_type_id, $bundle = NULL) {
    $connection = Database::getConnection('default', $database);
    $id_column = $this->getIdColumn($entity_type_id);
    $query = $connection
      ->select($entity_type_id, 'e')
      ->fields('e', [$id_column])
      ->range($offset, $length)
      ->orderBy("e.$id_column");
    if (!is_null($bundle)) {
      $query->condition('e.type', $bundle);
    }
    return $query
      ->execute()
      ->fetchCol();
  }

  /**
   * Return the ID column depending on the entity type.
   *
   * @param string $type
   *   Type of the entity.
   *
   * @return string
   *   ID column.
   */
  protected function getIdColumn($type) {
    return ($type === 'users') ? 'uid' : (($type === 'node') ? 'nid' : 'id');
  }

  /**
   * Check which entity ids are not in a database.
   *
   * @param array $ids
   *   List of entity ids to check.
   * @param string $database
   *   Database to check.
   * @param string $entity_type_id
   *   Type of the entities.
   * @param string $bundle
   *   (Optional) Bundle of the entity type.
   *
   * @return array
   *   Array of entity ids not present in the given database.
   */
  protected function entityIdsNotInDatabase($ids, $database, $entity_type_id, $bundle = NULL) {
    $connection = Database::getConnection('default', $database);
    $id_column = $this->getIdColumn($entity_type_id);
    $query = $connection
      ->select($entity_type_id, 'e')
      ->fields('e', [$id_column])
      ->condition("e.$id_column", $ids, 'IN')
      ->orderBy("e.$id_column");
    if (!is_null($bundle)) {
      $query->condition('e.type', $bundle);
    }
    $matched_ids = $query
      ->execute()
      ->fetchCol();

    return array_diff($ids, $matched_ids);
  }

}
